/*******************************************************************************
 (c) 2005-2014 Copyright, Real-Time Innovations, Inc.  All rights reserved.
 RTI grants Licensee a license to use, modify, compile, and create derivative
 works of the Software.  Licensee has the right to distribute object form only
 for use with RTI products.  The Software is provided "as is", with no warranty
 of any type, including any warranty for fitness for any purpose. RTI is under
 no obligation to maintain or support the Software.  RTI shall not be liable for
 any incidental or consequential damages arising out of the use or inability to
 use the software.
 ******************************************************************************/
/* cfc_publisher.c

   A publication of data of type cfc

   This file is derived from code automatically generated by the rtiddsgen
   command:

   rtiddsgen -language C -example <arch> cfc.idl

   Example publication of type cfc automatically generated by
   'rtiddsgen'. To test them follow these steps:

   (1) Compile this file and the example subscription.

   (2) Start the subscription with the command
       objs/<arch>/cfc_subscriber <domain_id> <sample_count>

   (3) Start the publication with the command
       objs/<arch>/cfc_publisher <domain_id> <sample_count>

   (4) [Optional] Specify the list of discovery initial peers and
       multicast receive addresses via an environment variable or a file
       (in the current working directory) called NDDS_DISCOVERY_PEERS.

   You can run any number of publishers and subscribers programs, and can
   add and remove them dynamically from the domain.


   Example:

       To run the example application on domain <domain_id>:

       On Unix:

       objs/<arch>/cfc_publisher <domain_id>
       objs/<arch>/cfc_subscriber <domain_id>

       On Windows:

       objs\<arch>\cfc_publisher <domain_id>
       objs\<arch>\cfc_subscriber <domain_id>


modification history
------------ -------
*/

#include "cfc.h"
#include "cfcSupport.h"
#include "ndds/ndds_c.h"
#include <stdio.h>
#include <stdlib.h>

/* Delete all entities */
static int publisher_shutdown(
        DDS_DomainParticipant *participant,
        struct DDS_DomainParticipantQos *participant_qos,
        struct DDS_DataWriterQos *datawriter_qos)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    if (participant != NULL) {
        retcode = DDS_DomainParticipant_delete_contained_entities(participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDS_DomainParticipantFactory_delete_participant(
                DDS_TheParticipantFactory,
                participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    retcode = DDS_DomainParticipantQos_finalize(participant_qos);
    if (retcode != DDS_RETCODE_OK) {
        printf("participantQos_finalize error %d\n", retcode);
        status = -1;
    }

    retcode = DDS_DataWriterQos_finalize(datawriter_qos);
    if (retcode != DDS_RETCODE_OK) {
        printf("dataWriterQos_finalize error %d\n", retcode);
        status = -1;
    }

    /* RTI Connext provides finalize_instance() method on
       domain participant factory for people who want to release memory used
       by the participant factory. Uncomment the following block of code for
       clean destruction of the singleton. */
    /*
        retcode = DDS_DomainParticipantFactory_finalize_instance();
        if (retcode != DDS_RETCODE_OK) {
            printf("finalize_instance error %d\n", retcode);
            status = -1;
        }
    */

    return status;
}

static int publisher_main(int domainId, int sample_count)
{
    DDS_DomainParticipant *participant = NULL;
    DDS_Publisher *publisher = NULL;
    DDS_Topic *topic = NULL;
    DDS_DataWriter *writer = NULL;
    cfcDataWriter *cfc_writer = NULL;
    cfc *instance = NULL;
    DDS_ReturnCode_t retcode;
    DDS_InstanceHandle_t instance_handle = DDS_HANDLE_NIL;
    const char *type_name = NULL;
    int count = 0;
    int i;
    int sample;
    struct DDS_Duration_t send_period = { 1, 0 };
    const char *cfc_name = "custom_flowcontroller";
    struct DDS_FlowControllerProperty_t custom_fcp;
    DDS_FlowController *cfc = NULL;
    struct DDS_DataWriterQos datawriter_qos = DDS_DataWriterQos_INITIALIZER;
    struct DDS_DomainParticipantQos participant_qos =
            DDS_DomainParticipantQos_INITIALIZER;

    participant = DDS_DomainParticipantFactory_create_participant(
            DDS_TheParticipantFactory,
            domainId,
            &DDS_PARTICIPANT_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        printf("create_participant error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* Start changes for custom_flowcontroller */

    /* If you want to change the Participant's QoS programmatically rather than
     * using the XML file, you will need to add the following lines to your
     * code and comment out the create_participant call above.
     */

    /* Get default participant QoS to customize */
    /*    retcode = DDS_DomainParticipantFactory_get_default_participant_qos(
                DDS_TheParticipantFactory, &participant_qos);
        if (retcode != DDS_RETCODE_OK) {
            printf("get_default_participant_qos error\n");
            return -1;
        }

    */    /* By default, data will be sent via shared memory _and_ UDPv4.  Because
     * the flowcontroller limits writes across all interfaces, this halves the
     * effective send rate.  To avoid this, we enable only the UDPv4 transport
     */
    /*    participant_qos.transport_builtin.mask = DDS_TRANSPORTBUILTIN_UDPv4;

    */    /* To create participant with default QoS, use DDS_PARTICIPANT_QOS_DEFAULT
       instead of participant_qos */
    /*    participant = DDS_DomainParticipantFactory_create_participant(
            DDS_TheParticipantFactory, domainId, &participant_qos,
            NULL, DDS_STATUS_MASK_NONE);
        if (participant == NULL) {
            printf("create_participant error\n");
           publisher_shutdown(participant, &participant_qos, &datawriter_qos);
            return -1;
        }
    */
    /* End changes for Custom_Flowcontroller */

    /* To customize publisher QoS, use
     the configuration file USER_QOS_PROFILES.xml */
    publisher = DDS_DomainParticipant_create_publisher(
            participant,
            &DDS_PUBLISHER_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (publisher == NULL) {
        printf("create_publisher error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* Register type before creating topic */
    type_name = cfcTypeSupport_get_type_name();
    retcode = cfcTypeSupport_register_type(participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        printf("register_type error %d\n", retcode);
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* To customize topic QoS, use
     the configuration file USER_QOS_PROFILES.xml */
    topic = DDS_DomainParticipant_create_topic(
            participant,
            "Example cfc",
            type_name,
            &DDS_TOPIC_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        printf("create_topic error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* To customize data writer QoS, use
     the configuration file USER_QOS_PROFILES.xml */
    writer = DDS_Publisher_create_datawriter(
            publisher,
            topic,
            &DDS_DATAWRITER_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (writer == NULL) {
        printf("create_datawriter error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* Start changes for custom_flowcontroller */

    /* If you want to change the Datawriter's QoS programmatically rather than
     * using the XML file, you will need to add the following lines to your
     * code and comment out the create_datawriter call above.
     *
     * In this case we create the flowcontroller and the neccesary QoS
     * for the datawriter.
     */

    /*    retcode = DDS_DomainParticipant_get_default_flowcontroller_property(
            participant, &custom_fcp);
        if (retcode != DDS_RETCODE_OK) {
            printf("get_default_flowcontroller_property error \n");
            return -1;
        }


    */    /* Don't allow too many tokens to accumulate */
    /*    custom_fcp.token_bucket.max_tokens =
            custom_fcp.token_bucket.tokens_added_per_period = 2;
        custom_fcp.token_bucket.tokens_leaked_per_period = DDS_LENGTH_UNLIMITED;

    */    /* 100ms */
    /*    custom_fcp.token_bucket.period.sec = 0;
        custom_fcp.token_bucket.period.nanosec = 100000000;

    */    /* The sample size is 1000, but the minimum bytes_per_token is 1024.
     * Furthermore, we want to allow some overhead.
     */
    /*    custom_fcp.token_bucket.bytes_per_token = 1024;

    */    /* So, in summary, each token can be used to send about one message,
     * and we get 2 tokens every 100ms, so this limits transmissions to
     * about 20 messages per second.
     */
    /* Create flowcontroller and set properties */
    /*    cfc = DDS_DomainParticipant_create_flowcontroller(
            participant, DDS_String_dup(cfc_name), &custom_fcp);
        if (cfc == NULL) {
            printf("create_flowcontroller error\n");
            return -1;
        }

    */    /* Get default datawriter QoS to customize */
    /*    retcode = DDS_Publisher_get_default_datawriter_qos(publisher,
                &datawriter_qos);
        if (retcode != DDS_RETCODE_OK) {
            printf("get_default_datawriter_qos error\n");
            return -1;
        }

    */    /* As an alternative to increasing history depth, we can just
     * set the qos to keep all samples
     */
    /*    datawriter_qos.history.kind = DDS_KEEP_ALL_HISTORY_QOS;
     */
    /* Set flowcontroller for datawriter */
    /*    datawriter_qos.publish_mode.kind = DDS_ASYNCHRONOUS_PUBLISH_MODE_QOS;
        datawriter_qos.publish_mode.flow_controller_name =
       DDS_String_dup(cfc_name);

    */    /* To create datawriter with default QoS, use DDS_DATAWRITER_QOS_DEFAULT
       instead of datawriter_qos */
    /*    writer = DDS_Publisher_create_datawriter(
            publisher, topic,
            &datawriter_qos, NULL, DDS_STATUS_MASK_NONE);
        if (writer == NULL) {
            printf("create_datawriter error\n");
           publisher_shutdown(participant, &participant_qos, &datawriter_qos);
            return -1;
        }
    */
    /* End changes for Custom_Flowcontroller */

    cfc_writer = cfcDataWriter_narrow(writer);
    if (cfc_writer == NULL) {
        printf("DataWriter narrow error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* Create data sample for writing */

    instance = cfcTypeSupport_create_data_ex(DDS_BOOLEAN_TRUE);

    if (instance == NULL) {
        printf("cfcTypeSupport_create_data error\n");
        publisher_shutdown(participant, &participant_qos, &datawriter_qos);
        return -1;
    }

    /* For a data type that has a key, if the same instance is going to be
       written multiple times, initialize the key here
       and register the keyed instance prior to writing */
    /*
        instance_handle = cfcDataWriter_register_instance(
            cfc_writer, instance);
    */

    /* Main loop */
    for (count = 0; (sample_count == 0) || (count < sample_count); ++count) {
        /* Changes for custom_flowcontroller */
        /* Simulate bursty writer */
        NDDS_Utility_sleep(&send_period);

        for (i = 0; i < 10; ++i) {
            sample = count * 10 + i;
            printf("Writing cfc, sample %d\n", sample);
            instance->x = sample;
            memset(instance->str, 1, 999);
            instance->str[999] = 0;
            retcode =
                    cfcDataWriter_write(cfc_writer, instance, &instance_handle);
            if (retcode != DDS_RETCODE_OK) {
                printf("write error %d\n", retcode);
            }
        }
    }

    /* This new sleep it is for let time to the subscriber to read all the
     * sent samples.
     */
    send_period.sec = 4;
    NDDS_Utility_sleep(&send_period);

    /*
        retcode = cfcDataWriter_unregister_instance(
            cfc_writer, instance, &instance_handle);
        if (retcode != DDS_RETCODE_OK) {
            printf("unregister instance error %d\n", retcode);
        }
    */

    /* Delete data sample */
    retcode = cfcTypeSupport_delete_data_ex(instance, DDS_BOOLEAN_TRUE);
    if (retcode != DDS_RETCODE_OK) {
        printf("cfcTypeSupport_delete_data error %d\n", retcode);
    }

    /* Cleanup and delete delete all entities */
    return publisher_shutdown(participant, &participant_qos, &datawriter_qos);
}

#if defined(RTI_WINCE)
int wmain(int argc, wchar_t **argv)
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = _wtoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = _wtoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDS_Config_Logger_set_verbosity_by_category(
        NDDS_Config_Logger_get_instance(),
        NDDS_CONFIG_LOG_CATEGORY_API,
        NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return publisher_main(domainId, sample_count);
}
#elif !(defined(RTI_VXWORKS) && !defined(__RTP__)) && !defined(RTI_PSOS)
int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDS_Config_Logger_set_verbosity_by_category(
        NDDS_Config_Logger_get_instance(),
        NDDS_CONFIG_LOG_CATEGORY_API,
        NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return publisher_main(domainId, sample_count);
}
#endif

#ifdef RTI_VX653
const unsigned char *__ctype = NULL;

void usrAppInit()
{
    #ifdef USER_APPL_INIT
    USER_APPL_INIT; /* for backwards compatibility */
    #endif

    /* add application specific code here */
    taskSpawn(
            "pub",
            RTI_OSAPI_THREAD_PRIORITY_NORMAL,
            0x8,
            0x150000,
            (FUNCPTR) publisher_main,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0);
}
#endif
